/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.wsdl.Definition;
import javax.wsdl.PortType;
import javax.xml.XMLConstants;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.components.JoinConditionComponent;
import org.ow2.orchestra.bpmn2bpel.components.LinkComponent;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpmn.Definitions;
import org.ow2.orchestra.jaxb.bpmn.Import;
import org.ow2.orchestra.jaxb.bpmn.MessageFlow;
import org.ow2.orchestra.jaxb.bpmn.Operation;
import org.ow2.orchestra.jaxb.bpmn.Participant;
import org.ow2.orchestra.jaxb.bpmn.TCollaboration;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TEndEvent;
import org.ow2.orchestra.jaxb.bpmn.TEndPoint;
import org.ow2.orchestra.jaxb.bpmn.TFlowElement;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TInterface;
import org.ow2.orchestra.jaxb.bpmn.TItemDefinition;
import org.ow2.orchestra.jaxb.bpmn.TMessage;
import org.ow2.orchestra.jaxb.bpmn.TProcess;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.jaxb.bpmn.TStartEvent;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.QNameComparator;
import org.ow2.orchestra.util.wsdl.WsdlUtil;
import org.xml.sax.SAXException;

/**
 * @author Guillaume Porcher
 */
public class BpmnMap {

  private static final Logger LOG = Logger.getLogger(BpmnMap.class.getName());

  public static final String BPMN_2_0_NS = "http://www.omg.org/spec/BPMN/20100524/MODEL";

  private final Map<QName, TProcess> processes = new TreeMap<QName, TProcess>(QNameComparator.INSTANCE);

  private final Map<QName, TStartEvent> startEvents = new TreeMap<QName, TStartEvent>(QNameComparator.INSTANCE);

  private final Map<QName, TEndEvent> endEvents = new TreeMap<QName, TEndEvent>(QNameComparator.INSTANCE);

  private final Map<QName, TFlowNode> flowNodes = new TreeMap<QName, TFlowNode>(QNameComparator.INSTANCE);

  private final Map<QName, TSequenceFlow> sequenceFlows = new TreeMap<QName, TSequenceFlow>(QNameComparator.INSTANCE);

  private final Map<QName, Operation> operations = new TreeMap<QName, Operation>(QNameComparator.INSTANCE);
  private final Map<QName, TInterface> interfaces = new TreeMap<QName, TInterface>(QNameComparator.INSTANCE);
  private final Map<QName, TEndPoint> endpoints = new TreeMap<QName, TEndPoint>(QNameComparator.INSTANCE);

  private final Map<QName, TDataObject> dataObjects = new TreeMap<QName, TDataObject>(QNameComparator.INSTANCE);

  private final Map<QName, Participant> participants = new TreeMap<QName, Participant>(QNameComparator.INSTANCE);

  private final List<MessageFlow> messageFlows = new ArrayList<MessageFlow>();

  private final Map<String, String> namespaceAliases = new TreeMap<String, String>();

  private final Map<QName, TItemDefinition> itemDefinitions = new TreeMap<QName, TItemDefinition>(QNameComparator.INSTANCE);

  private final Map<QName, TMessage> messages = new TreeMap<QName, TMessage>(QNameComparator.INSTANCE);

  private final List<LinkComponent> links = new ArrayList<LinkComponent>();
  private final Map<TFlowNode, JoinConditionComponent> joinComponents = new HashMap<TFlowNode, JoinConditionComponent>();

  private final Map<String, Definition> importedWsdlDefinitions = new TreeMap<String, Definition>();

  /**
   * @return the startEvents
   */
  public Map<QName, TStartEvent> getStartEvents() {
    return this.startEvents;
  }

  /**
   * @return the endEvents
   */
  public Map<QName, TEndEvent> getEndEvents() {
    return this.endEvents;
  }

  /**
   * @return the flowNodes
   */
  public Map<QName, TFlowNode> getFlowNodes() {
    return this.flowNodes;
  }

  /**
   * @return the sequenceFlows
   */
  public Map<QName, TSequenceFlow> getSequenceFlows() {
    return this.sequenceFlows;
  }

  public List<TSequenceFlow> getOutgoingSequenceFlows(final TFlowNode node) {
    final List<TSequenceFlow> result = new ArrayList<TSequenceFlow>();
    for (final QName seqQName : node.getOutgoings()) {
      final TSequenceFlow seqFlow = this.sequenceFlows.get(seqQName);
      if (seqFlow == null) {
        throw new TransformerException("SequenceFlow " + seqQName + " not found !");
      }
      result.add(seqFlow);
    }
    return result;
  }

  public List<TSequenceFlow> getIncomingSequenceFlows(final TFlowNode node) {
    final List<TSequenceFlow> result = new ArrayList<TSequenceFlow>();
    for (final QName seqQName : node.getIncomings()) {
      final TSequenceFlow seqFlow = this.sequenceFlows.get(seqQName);
      if (seqFlow == null) {
        throw new TransformerException("SequenceFlow " + seqQName + " not found !");
      }
      result.add(seqFlow);
    }
    return result;
  }

  /**
   * @return the processes
   */
  public Map<QName, TProcess> getProcesses() {
    return this.processes;
  }

  /**
   * @return the importedWsdlDefinitions
   */
  public Map<String, Definition> getImportedWsdlDefinitions() {
    return this.importedWsdlDefinitions;
  }

  /**
   * @return the interfaces
   */
  public Map<QName, TInterface> getInterfaces() {
    return this.interfaces;
  }

  /**
   * @return the operations
   */
  public Map<QName, Operation> getOperations() {
    return this.operations;
  }

  /**
   * @return the endpoints
   */
  public Map<QName, TEndPoint> getEndpoints() {
    return this.endpoints;
  }

  /**
   * @return the participants
   */
  public Map<QName, Participant> getParticipants() {
    return this.participants;
  }

  /**
   * @return the itemDefinitions
   */
  public Map<QName, TItemDefinition> getItemDefinitions() {
    return this.itemDefinitions;
  }

  /**
   * @return the messages
   */
  public Map<QName, TMessage> getMessages() {
    return this.messages;
  }

  /**
   * @return the dataObject
   */
  public Map<QName, TDataObject> getDataObjects() {
    return this.dataObjects;
  }

  /**
   * @return the messageFlows
   */
  public List<MessageFlow> getMessageFlows() {
    return this.messageFlows;
  }

  public Collection<MessageFlow> getOutgoingMessageFlows(final TFlowNode node, final QName processQName) {
    final Collection<MessageFlow> result = new ArrayList<MessageFlow>();
    for (final MessageFlow messageFlow : this.messageFlows) {
      if (messageFlow.getSourceRef().equals(new QName(processQName.getNamespaceURI(), node.getId()))) {
        result.add(messageFlow);
      }
    }
    return result;
  }

  public String getNamespaceAlias(final String namespace) {
    if (!this.namespaceAliases.containsKey(namespace)) {
      this.namespaceAliases.put(namespace, "ns" + this.namespaceAliases.size());
    }
    return this.namespaceAliases.get(namespace);
  }

  public String getPartnerLinkName(final QName participantQName, final QName interfaceQName) {
    final String prefix = this.getNamespaceAlias(participantQName.getNamespaceURI());
    return prefix + "-" + participantQName.getLocalPart() + "-" + this.getRoleName(interfaceQName);
  }

  public String getPartnerLinkTypeName(final QName participantQName) {
    return this.getNamespaceAlias(participantQName.getNamespaceURI()) + "-" + participantQName.getLocalPart() + "Type";
  }

  public String getRoleName(final QName interfaceQName) {
    return this.getNamespaceAlias(interfaceQName.getNamespaceURI()) + "-" + interfaceQName.getLocalPart();
  }

  /**
   * Create a structure containing bpmn informations used for the generation
   *
   * @param definitions the Bpmn parsed object
   * @throws IOException
   * @throws SAXException
   * @throws JAXBException
   */
  public static void getBpmnMap(
          final Definitions definitions,
          final BpmnMap bpmnMap,
          final URL baseUrl)
          throws IOException, JAXBException, SAXException {
    // process imports
    for (final Import tImport : definitions.getImports()) {
      if (tImport.getImportType().equals(BpmnMap.BPMN_2_0_NS)) {
        final String location = tImport.getLocation();

        // get importedBpmn
        final URL url = new URL(baseUrl, location);
        final InputStream importedBpmn = url.openStream();

        // parse bpmn definition
        final Definitions importedDefinitions = JaxbUtil.parseBPMNDefinitions(importedBpmn);
        BpmnMap.getBpmnMap(importedDefinitions, bpmnMap, baseUrl);
      } else if (tImport.getImportType().equals(WSDLGenerator.XMLNS_WSDL)) {
        // parse imported wsdlDefinition
        final String location = tImport.getLocation();
        final URL url = new URL(baseUrl, location);
        final Definition wsdlDefinition = WsdlUtil.readWsdl(url);
        final String importName = Misc.getHumanReadableId("wsdl") + ".wsdl";
        bpmnMap.getImportedWsdlDefinitions().put(importName, wsdlDefinition);
      } else if (tImport.getImportType().endsWith(XMLConstants.W3C_XML_SCHEMA_NS_URI)) {
        // parse imported schema
        // TODO
      }
    }

    // find process element in the definition
    final List<JAXBElement<? extends TRootElement>> elements = definitions.getRootElements();
    for (final JAXBElement<? extends TRootElement> element : elements) {
      if (element.getValue() instanceof TProcess) {
        final TProcess bpmnProcess = (TProcess) element.getValue();
        // add process
        final String processId = bpmnProcess.getId() == null ? bpmnProcess.getName() : bpmnProcess.getId();
        bpmnMap.getProcesses().put(new QName(definitions.getTargetNamespace(), processId), bpmnProcess);

        // get process informations
        for (final JAXBElement<? extends TFlowElement> flowElement : bpmnProcess.getFlowElements()) {
          if (flowElement.getValue() instanceof TFlowNode) {
            final TFlowNode flowNode = (TFlowNode) flowElement.getValue();
            bpmnMap.getFlowNodes().put(new QName(definitions.getTargetNamespace(), flowNode.getId()), flowNode);
            if (flowNode instanceof TStartEvent) {
              final TStartEvent startEvent = (TStartEvent) flowNode;
              bpmnMap.getStartEvents().put(new QName(definitions.getTargetNamespace(), startEvent.getId()), startEvent);
            }
            if (flowNode instanceof TEndEvent) {
              final TEndEvent endEvent = (TEndEvent) flowNode;
              bpmnMap.getEndEvents().put(new QName(definitions.getTargetNamespace(), endEvent.getId()), endEvent);
            }
          }
          if (flowElement.getValue() instanceof TSequenceFlow) {
            final TSequenceFlow sequenceFlow = (TSequenceFlow) flowElement.getValue();
            bpmnMap.getSequenceFlows().put(new QName(definitions.getTargetNamespace(), sequenceFlow.getId()), sequenceFlow);
          }
          if (flowElement.getValue() instanceof TDataObject) {
            final TDataObject dataObject = (TDataObject) flowElement.getValue();
            bpmnMap.getDataObjects().put(new QName(definitions.getTargetNamespace(), dataObject.getId()), dataObject);
          }
        }

      } else if (element.getValue() instanceof TInterface) {
        // get interface informations
        final TInterface tInterface = (TInterface) element.getValue();
        bpmnMap.getInterfaces().put(new QName(definitions.getTargetNamespace(), tInterface.getId()), tInterface);
        for (final Operation operation : tInterface.getOperations()) {
          final QName operationQName = new QName(definitions.getTargetNamespace(), operation.getId());
          bpmnMap.getOperations().put(operationQName, operation);
        }
      } else if (element.getValue() instanceof TEndPoint) {
        // get interface informations
        final TEndPoint tEndPoint = (TEndPoint) element.getValue();
        bpmnMap.getEndpoints().put(new QName(definitions.getTargetNamespace(), tEndPoint.getId()), tEndPoint);
      } else if (element.getValue() instanceof TCollaboration) {
        // get information about message flows and participants
        final TCollaboration collaboration = (TCollaboration) element.getValue();
        for (final Participant participant : collaboration.getParticipants()) {
          final QName participantQName = new QName(definitions.getTargetNamespace(), participant.getId());
          bpmnMap.getParticipants().put(participantQName, participant);
        }
        for (final MessageFlow messageFlow : collaboration.getMessageFlows()) {
          bpmnMap.getMessageFlows().add(messageFlow);
        }
      } else if (element.getValue() instanceof TItemDefinition) {
        // get item definitions
        final TItemDefinition itemDefinition = (TItemDefinition) element.getValue();
        final QName itemQName = new QName(definitions.getTargetNamespace(), itemDefinition.getId());
        bpmnMap.getItemDefinitions().put(itemQName, itemDefinition);
      } else if (element.getValue() instanceof TMessage) {
        // get message definitions
        final TMessage message = (TMessage) element.getValue();
        final QName messageQName = new QName(definitions.getTargetNamespace(), message.getId());
        bpmnMap.getMessages().put(messageQName, message);
      }
    }

    // fill incomings and outgoings for each activity.
    for (final TFlowNode node : bpmnMap.getFlowNodes().values()) {
      if (node.getIncomings().isEmpty()) {
        for (final Entry<QName, TSequenceFlow> entry : bpmnMap.getSequenceFlows().entrySet()) {
          if (entry.getValue().getTargetRef().equals(node)) {
            node.getIncomings().add(entry.getKey());
          }
        }
      }
      if (node.getOutgoings().isEmpty()) {
        for (final Entry<QName, TSequenceFlow> entry : bpmnMap.getSequenceFlows().entrySet()) {
          if (entry.getValue().getSourceRef().equals(node)) {
            node.getOutgoings().add(entry.getKey());
          }
        }
      }
    }
  }

  /**
   * @return the links
   */
  public List<LinkComponent> getLinks() {
    return this.links;
  }

  public JoinConditionComponent getJoinCondition(final TFlowNode node) {
    return this.joinComponents.get(node);
  }

  /**
   * @return the joinComponents
   */
  public Map<TFlowNode, JoinConditionComponent> getJoinComponents() {
    return this.joinComponents;
  }

  public List<LinkComponent> getIncomingLinks(final TFlowNode node) {
    final List<LinkComponent> result = new ArrayList<LinkComponent>();
    for (final LinkComponent link : this.links) {
      if (link.getTargetNode().equals(node)) {
        result.add(link);
      }
    }
    return result;
  }

  public List<LinkComponent> getOutgoingLinks(final TFlowNode node) {
    final List<LinkComponent> result = new ArrayList<LinkComponent>();
    for (final LinkComponent link : this.links) {
      if (link.getSourceNode().equals(node)) {
        result.add(link);
      }
    }
    return result;
  }

  public TSequenceFlow createSequenceFlow(final TFlowNode sourceNode, final TFlowNode targetNode) {
    final TSequenceFlow sequenceFlow = new TSequenceFlow();
    sequenceFlow.setSourceRef(sourceNode);
    sequenceFlow.setTargetRef(targetNode);
    final QName seqQName = new QName(Misc.getHumanReadableId("sequenceFlow"));
    sequenceFlow.setId(seqQName.getLocalPart());
    this.sequenceFlows.put(seqQName, sequenceFlow);
    sourceNode.getOutgoings().add(seqQName);
    targetNode.getIncomings().add(seqQName);
    Misc.fastDynamicLog(BpmnMap.LOG, Level.INFO, "Creating sequenceFlow %s: %s (%s) -> %s (%s)",
            sequenceFlow.getId(),
            sourceNode.getId(),
            sourceNode.getClass().getSimpleName(),
            targetNode.getId(),
            targetNode.getClass().getSimpleName());
    return sequenceFlow;
  }

  public void changeSequenceFlowTarget(final TSequenceFlow sequenceFlow, final TFlowNode targetNode) {
    final TFlowNode previousNode = (TFlowNode) sequenceFlow.getTargetRef();
    sequenceFlow.setTargetRef(targetNode);
    QName seqQName = null;
    for (final Entry<QName, TSequenceFlow> entry : this.sequenceFlows.entrySet()) {
      if (sequenceFlow.equals(entry.getValue())) {
        seqQName = entry.getKey();
      }
    }
    if (seqQName != null) {
      previousNode.getIncomings().remove(seqQName);
      targetNode.getIncomings().add(seqQName);
    } else {
      throw new TransformerException("sequenceFlow not found !");
    }
    Misc.fastDynamicLog(BpmnMap.LOG, Level.INFO, "Changing sequenceFlow %s target: %s (%s)",
            sequenceFlow.getId(),
            targetNode.getId(),
            targetNode.getClass().getSimpleName());
  }

  public void changeSequenceFlowSource(final TSequenceFlow sequenceFlow, final TFlowNode sourceNode) {
    final TFlowNode previousNode = (TFlowNode) sequenceFlow.getSourceRef();
    sequenceFlow.setSourceRef(sourceNode);
    QName seqQName = null;
    for (final Entry<QName, TSequenceFlow> entry : this.sequenceFlows.entrySet()) {
      if (sequenceFlow.equals(entry.getValue())) {
        seqQName = entry.getKey();
      }
    }
    if (seqQName != null) {
      previousNode.getOutgoings().remove(seqQName);
      sourceNode.getOutgoings().add(seqQName);
    } else {
      throw new TransformerException("sequenceFlow not found !");
    }
    Misc.fastDynamicLog(BpmnMap.LOG, Level.INFO, "Changing sequenceFlow %s source: %s (%s)",
            sequenceFlow.getId(),
            sourceNode.getId(),
            sourceNode.getClass().getSimpleName());
  }

  public void deleteSequenceFlow(final TSequenceFlow sequenceFlow) {
    QName seqQName = null;
    for (final Entry<QName, TSequenceFlow> entry : this.sequenceFlows.entrySet()) {
      if (sequenceFlow.equals(entry.getValue())) {
        seqQName = entry.getKey();
      }
    }
    if (seqQName != null) {
      this.sequenceFlows.remove(seqQName);
      ((TFlowNode) sequenceFlow.getSourceRef()).getOutgoings().remove(seqQName);
      ((TFlowNode) sequenceFlow.getTargetRef()).getIncomings().remove(seqQName);
    } else {
      throw new TransformerException("sequenceFlow not found !");
    }
    Misc.fastDynamicLog(BpmnMap.LOG, Level.INFO, "Removing sequenceFlow %s: %s (%s) -> %s (%s)",
            sequenceFlow.getId(),
            ((TFlowNode) sequenceFlow.getSourceRef()).getId(),
            ((TFlowNode) sequenceFlow.getSourceRef()).getClass().getSimpleName(),
            ((TFlowNode) sequenceFlow.getTargetRef()).getId(),
            ((TFlowNode) sequenceFlow.getTargetRef()).getClass().getSimpleName());
  }


  public void removeNode(final TFlowNode node) {
    this.getFlowNodes().values().remove(node);
    for (final TSequenceFlow seqFlow : this.getIncomingSequenceFlows(node)) {
      this.deleteSequenceFlow(seqFlow);
    }
    for (final TSequenceFlow seqFlow : this.getOutgoingSequenceFlows(node)) {
      this.deleteSequenceFlow(seqFlow);
    }
  }


  public PortType getPortType(QName portTypeQName) {
    for (Definition wsdlDefinition : getImportedWsdlDefinitions().values()) {
      PortType pt = wsdlDefinition.getPortType(portTypeQName);
      if (pt != null) {
        return pt;
      }
    }
    return null;
  }
}
