/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel;

import java.io.ByteArrayOutputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.Map.Entry;

import javax.wsdl.Binding;
import javax.wsdl.BindingInput;
import javax.wsdl.BindingOperation;
import javax.wsdl.BindingOutput;
import javax.wsdl.Definition;
import javax.wsdl.Import;
import javax.wsdl.Input;
import javax.wsdl.Message;
import javax.wsdl.Output;
import javax.wsdl.Part;
import javax.wsdl.Port;
import javax.wsdl.PortType;
import javax.wsdl.Service;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.UnknownExtensibilityElement;
import javax.wsdl.extensions.schema.Schema;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.wsdl.extensions.soap.SOAPBinding;
import javax.wsdl.extensions.soap.SOAPBody;
import javax.wsdl.extensions.soap.SOAPOperation;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.wsdl.xml.WSDLWriter;
import javax.xml.XMLConstants;
import javax.xml.namespace.QName;

import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpmn.Operation;
import org.ow2.orchestra.jaxb.bpmn.Participant;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TEndPoint;
import org.ow2.orchestra.jaxb.bpmn.TInterface;
import org.ow2.orchestra.jaxb.bpmn.TItemDefinition;
import org.ow2.orchestra.jaxb.bpmn.TMessage;
import org.ow2.orchestra.jaxb.bpmn.TProcess;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.ibm.wsdl.extensions.schema.SchemaConstants;
import com.ibm.wsdl.extensions.schema.SchemaImpl;
import com.ibm.wsdl.extensions.soap.SOAPAddressImpl;
import com.ibm.wsdl.extensions.soap.SOAPBindingImpl;
import com.ibm.wsdl.extensions.soap.SOAPBodyImpl;
import com.ibm.wsdl.extensions.soap.SOAPOperationImpl;


/**
 * @author Guillaume Porcher
 */
public final class WSDLGenerator {

  /**
   * Orchestra constant: replaced in orchestra engine with the server url.
   */
  private static final String ORCHESTRA_WS_URL = "${orchestra.ws.url}/";
  /**
   *
   */
  private static final String MSG_PART_NAME = "payload";

  private WSDLGenerator() {
  }

  /**
   * ws-addressing ns
   */
  public static final String XMLNS_WS_ADDRESSING = "http://www.w3.org/2005/08/addressing";
  /**
   * bpel 2.0 partnerLinkType ns.
   */
  public static final String XMLNS_BPEL_2_0_PLNKTYPE = "http://docs.oasis-open.org/wsbpel/2.0/plnktype";
  /**
   * wsdl ns
   */
  public static final String XMLNS_WSDL = "http://schemas.xmlsoap.org/wsdl/";
  public static final String XMLNS_WSDL_SOAP = "http://schemas.xmlsoap.org/wsdl/soap/";

  private static final WSDLFactory WSDL_FACTORY;

  static {
    try {
      WSDL_FACTORY = WSDLFactory.newInstance();
    }
    catch (final WSDLException e) {
      throw new RuntimeException(e);
    }
  }

  public static Map<String, Definition> generateWsdlDefinitions(
          final BpmnMap bpmnMap, final QName processQName) throws WSDLException {
    final Definition wsdlDefinition = WSDLGenerator.WSDL_FACTORY.newDefinition();
    wsdlDefinition.setQName(processQName);
    wsdlDefinition.setTargetNamespace(processQName.getNamespaceURI());
    wsdlDefinition.addNamespace("tns", processQName.getNamespaceURI());
    wsdlDefinition.addNamespace("xsd", XMLConstants.W3C_XML_SCHEMA_NS_URI);
    wsdlDefinition.addNamespace("soap", WSDLGenerator.XMLNS_WSDL_SOAP);
    wsdlDefinition.setTypes(wsdlDefinition.createTypes());

    // get partner wsdl definitions
    final Map<String, Definition> partnerWSDLs = WSDLGenerator.getPartnerWSDLs(bpmnMap);

    // add imported wsdl definitions
    partnerWSDLs.putAll(WSDLGenerator.getImportedWSDLs(bpmnMap));

    if (!bpmnMap.getParticipants().isEmpty()) {

      // generate WSDL port types for the process
      WSDLGenerator.generateProcessWSDL(bpmnMap, processQName, wsdlDefinition);

      // Add wsdl and bpel imports
      for (final Entry<String, Definition> partnerWSDL : partnerWSDLs.entrySet()) {
        final javax.wsdl.Import wsdlImport = wsdlDefinition.createImport();
        wsdlImport.setDefinition(partnerWSDL.getValue());
        wsdlImport.setNamespaceURI(partnerWSDL.getValue().getTargetNamespace());
        wsdlImport.setLocationURI(partnerWSDL.getKey());
        wsdlDefinition.addImport(wsdlImport);
      }

      // create partnerLinkType for participants
      WSDLGenerator.addPartnerLinkTypes(bpmnMap, wsdlDefinition);

    }

    partnerWSDLs.put("process.wsdl", wsdlDefinition);

    return partnerWSDLs;
  }

  /**
   * Get a map of WSDL definitions for definitions embedded in a endpoint.
   *
   * @param bpmnMap
   * @return
   * @throws WSDLException
   */
  private static Map<String, Definition> getPartnerWSDLs(final BpmnMap bpmnMap) throws WSDLException {
    final Map<String, Definition> partnerWsdls = new TreeMap<String, Definition>();
    for (final Participant participant : bpmnMap.getParticipants().values()) {
      for (final QName endPointQName : participant.getEndPointReves()) {
        final TEndPoint endPoint = bpmnMap.getEndpoints().get(endPointQName);
        final List<Element> endPointElements = endPoint.getExtensionElements().getAnies();
        for (final Element element : endPointElements) {
          if (element.getNamespaceURI().equals(WSDLGenerator.XMLNS_WS_ADDRESSING)
                  && element.getLocalName().equals("EndpointReference")) {
            final Element metadataElt = XmlUtil.element(element, WSDLGenerator.XMLNS_WS_ADDRESSING, "Metadata");
            if (metadataElt != null) {
              final Element wsdlElt = XmlUtil.element(metadataElt, WSDLGenerator.XMLNS_WSDL, "definitions");
              if (wsdlElt != null) {
                final WSDLFactory wsdlFactory = WSDLFactory.newInstance();
                final WSDLReader wsdlReader = wsdlFactory.newWSDLReader();
                final Definition definition = wsdlReader.readWSDL((String) null, wsdlElt);
                partnerWsdls.put("partner" + partnerWsdls.size() + ".wsdl", definition);
              }
            }
          }
        }
      }
    }
    return partnerWsdls;
  }

  /**
   * Get a map of WSDL definitions for definitions imported by the BPMN file.
   *
   * @param bpmnMap
   * @return
   * @throws WSDLException
   */
  private static Map<String, Definition> getImportedWSDLs(final BpmnMap bpmnMap) throws WSDLException {
    return bpmnMap.getImportedWsdlDefinitions();
  }

  private static void generateProcessWSDL(final BpmnMap bpmnMap, final QName processQName, final Definition wsdlDefinition) {
    final Set<QName> itemDefinitionStructuresToAdd = new HashSet<QName>();

    for (final Entry<QName, Participant> entry : bpmnMap.getParticipants().entrySet()) {
      final Participant participant = entry.getValue();
      if (processQName.equals(participant.getProcessRef())) {
        // Process participant
        // add wsdl informations
        for (final QName interfaceQName : participant.getInterfaceReves()) {
          final TInterface tInterface = bpmnMap.getInterfaces().get(interfaceQName);

          // generate PortType
          final PortType portType = wsdlDefinition.createPortType();
          portType.setQName(new QName(interfaceQName.getNamespaceURI(), tInterface.getName(), interfaceQName.getPrefix()));
          portType.setUndefined(false);

          wsdlDefinition.addPortType(portType);

          // add soap binding for the PortType
          final Binding binding = wsdlDefinition.createBinding();
          binding.setQName(
                  new QName(
                          interfaceQName.getNamespaceURI(),
                          tInterface.getName() + "SOAPBinding",
                          interfaceQName.getPrefix()));
          binding.setUndefined(false);
          binding.setPortType(portType);
          final SOAPBinding soapBinding = new SOAPBindingImpl();
          soapBinding.setStyle("document");
          soapBinding.setTransportURI("http://schemas.xmlsoap.org/soap/http");
          binding.addExtensibilityElement(soapBinding);

          wsdlDefinition.addBinding(binding);

          for (final Operation op : tInterface.getOperations()) {
            // add operations to portType
            final javax.wsdl.Operation operation = wsdlDefinition.createOperation();
            operation.setName(op.getName());
            operation.setUndefined(false);

            portType.addOperation(operation);

            // add binding for operation
            final BindingOperation bindingOperation = wsdlDefinition.createBindingOperation();
            bindingOperation.setName(op.getName());
            bindingOperation.setOperation(operation);
            binding.addBindingOperation(bindingOperation);
            final SOAPOperation soapOperation = new SOAPOperationImpl();
            soapOperation.setSoapActionURI(processQName.getNamespaceURI() + processQName.getLocalPart() + "/" + op.getName());
            bindingOperation.addExtensibilityElement(soapOperation);

            final QName inMessageQName = op.getInMessageRef();
            // generate inMessage
            final Message message = WSDLGenerator.generateWSDLMessage(bpmnMap, wsdlDefinition, inMessageQName);
            itemDefinitionStructuresToAdd.add(bpmnMap.getMessages().get(inMessageQName).getItemRef());

            final Input input = wsdlDefinition.createInput();
            input.setMessage(message);
            operation.setInput(input);

            final BindingInput bindingInput = wsdlDefinition.createBindingInput();
            final SOAPBody inputSoapBody = new SOAPBodyImpl();
            inputSoapBody.setUse("literal");
            bindingInput.addExtensibilityElement(inputSoapBody);
            bindingOperation.setBindingInput(bindingInput);

            // generate outMessage
            if (op.getOutMessageRef() != null) {
              final QName outMessageQName = op.getOutMessageRef();
              final Message outWsdlMessage = WSDLGenerator.generateWSDLMessage(bpmnMap, wsdlDefinition, outMessageQName);
              itemDefinitionStructuresToAdd.add(bpmnMap.getMessages().get(outMessageQName).getItemRef());

              final Output output = wsdlDefinition.createOutput();
              output.setMessage(outWsdlMessage);
              operation.setOutput(output);

              final BindingOutput bindingOutput = wsdlDefinition.createBindingOutput();
              final SOAPBody outputSoapBody = new SOAPBodyImpl();
              outputSoapBody.setUse("literal");
              bindingOutput.addExtensibilityElement(outputSoapBody);
              bindingOperation.setBindingOutput(bindingOutput);
            }
          }
        }
      }
    }
    // add service and ports
    final Service service = wsdlDefinition.createService();
    final TProcess tProcess = bpmnMap.getProcesses().get(processQName);
    service.setQName(new QName(
            processQName.getNamespaceURI(),
            tProcess.getName() + "Service",
            processQName.getPrefix()));
    wsdlDefinition.addService(service);

    for (final Object bindingEntry : wsdlDefinition.getBindings().entrySet()) {
      final Binding binding = ((Entry<QName, Binding>) bindingEntry).getValue();
      final Port port = wsdlDefinition.createPort();
      port.setBinding(binding);
      port.setName(binding.getPortType().getQName().getLocalPart() + "Port");

      service.addPort(port);

      final SOAPAddress soapAddress = new SOAPAddressImpl();
      port.addExtensibilityElement(soapAddress);
      soapAddress.setLocationURI(WSDLGenerator.ORCHESTRA_WS_URL + tProcess.getName() + "/" + port.getName());
    }
    // add dataObject xml types to wsdl
    for (final TDataObject dataObject : bpmnMap.getDataObjects().values()) {
      final QName itemDefinitionQName = dataObject.getItemSubjectRef();
      itemDefinitionStructuresToAdd.add(itemDefinitionQName);
    }
    // generate wsdl types
    for (final QName itemDefinitionQName : itemDefinitionStructuresToAdd) {
      WSDLGenerator.addItemDefinitionXsdType(bpmnMap, wsdlDefinition, itemDefinitionQName);
    }
  }

  private static Message generateWSDLMessage(final BpmnMap bpmnMap, final Definition wsdlDefinition, final QName messageQName) {
    final TMessage inMessage = bpmnMap.getMessages().get(messageQName);

    final Message message = wsdlDefinition.createMessage();
    message.setUndefined(false);
    message.setQName(new QName(messageQName.getNamespaceURI(), inMessage.getName(), messageQName.getPrefix()));

    final Part messagePart = wsdlDefinition.createPart();
    // Process will use Document/literal binding
    // According to WS-Basic profile 1.1, messages contains only one part.
    messagePart.setName(WSDLGenerator.MSG_PART_NAME);

    final QName itemDefinitionQName = inMessage.getItemRef();

    final TItemDefinition itemDefinition = bpmnMap.getItemDefinitions().get(itemDefinitionQName);
    if (itemDefinition == null) {
      throw new TransformerException("Missing itemDefinition: " + itemDefinitionQName);
    }
    if (itemDefinition.getStructureRef() != null) {
      messagePart.setTypeName(itemDefinition.getStructureRef());
    } else {
      final List<Element> outItemElements = itemDefinition.getExtensionElements().getAnies();
      boolean typeFound = false;
      for (final Element elt : outItemElements) {
        if (elt.getNamespaceURI().equals(XMLConstants.W3C_XML_SCHEMA_NS_URI)) {
          if (elt.getLocalName().equals("element")) {
            messagePart.setElementName(new QName(wsdlDefinition.getTargetNamespace(), elt.getAttribute("name")));
          } else {
            messagePart.setTypeName(new QName(wsdlDefinition.getTargetNamespace(), elt.getAttribute("name")));
          }
          if (typeFound) {
            throw new TransformerException(
                    "Invalid type definition for itemDefinition "
                            + itemDefinitionQName + ". Only one element or complexType is supported.");
          }
          typeFound = true;
        }
      }
    }

    message.addPart(messagePart);
    wsdlDefinition.addMessage(message);
    return message;
  }

  private static void addItemDefinitionXsdType(
          final BpmnMap bpmnMap,
          final Definition wsdlDefinition,
          final QName itemDefinitionQName) {
    final TItemDefinition itemDefinition = bpmnMap.getItemDefinitions().get(itemDefinitionQName);
    if (itemDefinition == null) {
      throw new TransformerException("Missing itemDefinition: " + itemDefinitionQName);
    }
    if (itemDefinition.getStructureRef() == null) {
      final List<Element> outItemElements = itemDefinition.getExtensionElements().getAnies();
      for (final Element elt : outItemElements) {
        if (elt.getNamespaceURI().equals(XMLConstants.W3C_XML_SCHEMA_NS_URI)) {
          final Element schemaElt = XmlUtil.getDocumentWithOneElement(SchemaConstants.Q_ELEM_XSD_2001);
          schemaElt.setPrefix("xsd");
          schemaElt.setAttribute("targetNamespace", itemDefinitionQName.getNamespaceURI());

          schemaElt.appendChild(schemaElt.getOwnerDocument().importNode(elt, true));

          final Schema schema = new SchemaImpl();
          schema.setElementType(SchemaConstants.Q_ELEM_XSD_2001);
          schema.setElement(schemaElt);
          wsdlDefinition.getTypes().addExtensibilityElement(schema);
        }
      }
    }
  }

  private static void addPartnerLinkTypes(final BpmnMap bpmnMap, final Definition wsdlDefinition) {
    for (final Entry<QName, Participant> entry : bpmnMap.getParticipants().entrySet()) {
      final QName participantQName = entry.getKey();
      final Participant participant = entry.getValue();

      // add partnerlinktype to wsdl
      final UnknownExtensibilityElement extElement = new UnknownExtensibilityElement();
      final Document document = XmlUtil.getNewDocument();
      final Element partnerLinkElt = document.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_PLNKTYPE, "plnk:partnerLinkType");
      document.appendChild(partnerLinkElt);
      partnerLinkElt.setAttribute("name", bpmnMap.getPartnerLinkTypeName(participantQName));
      extElement.setElement(partnerLinkElt);
      wsdlDefinition.addExtensibilityElement(extElement);

      for (final QName interfaceQName : participant.getInterfaceReves()) {
        final TInterface tInterface = bpmnMap.getInterfaces().get(interfaceQName);
        final Element roleElt = document.createElementNS(WSDLGenerator.XMLNS_BPEL_2_0_PLNKTYPE, "plnk:role");
        partnerLinkElt.appendChild(roleElt);
        roleElt.setAttribute("name", bpmnMap.getRoleName(interfaceQName));
        if (tInterface.getImplementationRef() != null) {
          roleElt.setAttribute(
              "portType",
              bpmnMap.getNamespaceAlias(tInterface.getImplementationRef().getNamespaceURI())
              + ":" + tInterface.getImplementationRef().getLocalPart());
        roleElt.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
              "xmlns:" + bpmnMap.getNamespaceAlias(tInterface.getImplementationRef().getNamespaceURI()),
              tInterface.getImplementationRef().getNamespaceURI());
        } else {
          roleElt.setAttribute(
                "portType",
                bpmnMap.getNamespaceAlias(interfaceQName.getNamespaceURI()) + ":" + tInterface.getName());
          roleElt.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
                "xmlns:" + bpmnMap.getNamespaceAlias(interfaceQName.getNamespaceURI()),
                interfaceQName.getNamespaceURI());
        }
      }
    }
  }

  /**
   * Recursively write the list of WSDL definitions.
   *
   * @param partnerWSDLs the list of WSDL definitions to write, with the wanted file name.
   *                     Imported definitions will be generated too, with generated names.
   * @return a map of the wsdls generated with file name as key and byte content as value.
   * @throws WSDLException
   */
  public static Map<String, byte[]> writeWsdls(final Map<String, Definition> partnerWSDLs) throws WSDLException {
    final Map<String, byte[]> files = new TreeMap<String, byte[]>();

    final WSDLWriter wsdlWriter = WSDLGenerator.WSDL_FACTORY.newWSDLWriter();

    final Map<Definition, String> writtenWsdls = new HashMap<Definition, String>();
    for (final Definition partnerWSDL : partnerWSDLs.values()) {
      files.putAll(WSDLGenerator.writeWsdl(partnerWSDL, partnerWSDLs, writtenWsdls, wsdlWriter));
    }
    return files;
  }

  /**
   * Recursively write a WSDL definition.
   *
   * @param wsdlToWrite  the main wsdl to write
   * @param partnerWSDLs a map of wsdl with fixed names (wsdl directly used by the process).
   *                     For definitions in this map, the key will be used as wsdl file name.
   *                     For definitions not in this map, a new name will be generated.
   * @param writtenWsdls a map of the wsdl definitions already written, with the file name associated.
   * @param wsdlWriter   the writer used to serialize definitions to XML.
   * @return a map of the wsdls generated with file name as key and byte content as value.
   * @throws WSDLException
   */
  private static Map<String, byte[]> writeWsdl(
          final Definition wsdlToWrite,
          final Map<String, Definition> partnerWSDLs,
          final Map<Definition, String> writtenWsdls,
          final WSDLWriter wsdlWriter) throws WSDLException {
    // do nothing if definition already generated
    if (writtenWsdls.containsKey(wsdlToWrite)) {
      return Collections.emptyMap();
    }

    final Map<String, byte[]> result = new TreeMap<String, byte[]>();
    // get name for this definition
    final String wsdlName;
    if (partnerWSDLs.containsValue(wsdlToWrite)) {
      String name = null;
      for (final Map.Entry<String, Definition> entry : partnerWSDLs.entrySet()) {
        if (entry.getValue().equals(wsdlToWrite)) {
          name = entry.getKey();
          break;
        }
      }
      wsdlName = name;
    } else {
      wsdlName = Misc.getHumanReadableId("wsdl") + ".wsdl";
    }
    // add to the list of wsdl definitions processed
    writtenWsdls.put(wsdlToWrite, wsdlName);

    // recursively write imported definitions
    // and update location
    for (final Collection<Import> wsdlImportVector : (Collection<Collection<Import>>) wsdlToWrite.getImports().values()) {
      for (final Import wsdlImport : wsdlImportVector) {
        final Definition importedDefinition = wsdlImport.getDefinition();
        if (!writtenWsdls.containsKey(importedDefinition)) {
          result.putAll(WSDLGenerator.writeWsdl(importedDefinition, partnerWSDLs, writtenWsdls, wsdlWriter));
        }
        wsdlImport.setLocationURI(writtenWsdls.get(importedDefinition));
      }
    }

    // generate xml for this definition
    final ByteArrayOutputStream partnerWsdlOutput = new ByteArrayOutputStream();

    wsdlWriter.writeWSDL(wsdlToWrite, partnerWsdlOutput);

    // Indent WSDL output
    final String indentedOut = XmlUtil.toString(XmlUtil.getDocumentFromString(partnerWsdlOutput.toString(), false));

    // Add content to list of files
    result.put(wsdlName, indentedOut.getBytes());

    return result;
  }

}
