/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.transformers;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.components.EmptyComponent;
import org.ow2.orchestra.bpmn2bpel.components.JoinConditionComponent;
import org.ow2.orchestra.bpmn2bpel.components.LinkComponent;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpmn.TEndEvent;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TGateway;
import org.ow2.orchestra.jaxb.bpmn.TParallelGateway;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.util.Misc;


/**
 * @author Guillaume Porcher
 */
public final class FlowLinkTransformer extends AbstractFlowLinkTransformer {

  private static final Logger LOG = Logger.getLogger(FlowLinkTransformer.class.getName());

  public boolean replaceFlow(final BpmnMap bpmnMap) {
    Misc.fastDynamicLog(FlowLinkTransformer.LOG, Level.INFO, "Trying to create link in a flow");
    // get join gateway
    for (final TFlowNode node : bpmnMap.getFlowNodes().values()) {
      if (node instanceof TParallelGateway) {
        final List<TSequenceFlow> incomingSequenceFlows = bpmnMap.getIncomingSequenceFlows(node);
        if (incomingSequenceFlows.size() > 1) {
          // node is a join gateway

          // try to replace sequenceFlow starting from a parallel gateway with a link
          for (final TSequenceFlow inSequenceFlow : incomingSequenceFlows) {
            final Object inSource = inSequenceFlow.getSourceRef();
            if ((inSource instanceof TParallelGateway)
                    && this.replaceIncomingSequenceFlowWithLink(inSequenceFlow, bpmnMap)) {
              // sequenceFlow replaced with a link
              return true;
            }
          }
          // replace first sequenceFlow going to node with a link

          // if node is the join gateway before the end event,
          // don't remove an incoming sequenceFlow from an activity
          // because the same transition would be re-created by the transformation.
          if (!(bpmnMap.getOutgoingSequenceFlows(node).get(0).getTargetRef() instanceof TEndEvent)) {
            if (this.replaceIncomingSequenceFlowWithLink(incomingSequenceFlows.get(0), bpmnMap)) {
              return true;
            }
          }
        }
      }
    }
    return false;
  }

  private boolean replaceIncomingSequenceFlowWithLink(final TSequenceFlow sequenceToReplace, final BpmnMap bpmnMap) {
    final TFlowNode gateway = (TFlowNode) sequenceToReplace.getTargetRef();
    final TFlowNode sourceNode = (TFlowNode) sequenceToReplace.getSourceRef();

    final List<TSequenceFlow> outSequences = bpmnMap.getOutgoingSequenceFlows(gateway);
    if (outSequences.size() != 1) {
      throw new TransformerException("Bidirectionnal gateways not supported");
    }
    final TFlowNode targetNode = (TFlowNode) outSequences.get(0).getTargetRef();

    TFlowNode linkSource = sourceNode;
    // if transition between a fork and a join gateways, replace with previous and next nodes
    if (sourceNode instanceof TParallelGateway) {
      final List<TSequenceFlow> sourceIncomingSequences = bpmnMap.getIncomingSequenceFlows(sourceNode);
      if (sourceIncomingSequences.size() == 1) {
        linkSource = (TFlowNode) sourceIncomingSequences.get(0).getSourceRef();
      } else {
        // if transition with another gateway
        // don't do transformation
        return false;
      }
    } else if (sourceNode instanceof TGateway) {
      // if transition with another gateway
      // don't do transformation
      return false;
    }
    if (targetNode instanceof TEndEvent) {
      // don't create link to end activity
      return false;
    } else if (targetNode instanceof TGateway) {
      // don't create link to a gateway, insert empty activity
      final EmptyComponent emptyComponent = new EmptyComponent();
      emptyComponent.setId(Misc.getHumanReadableId("component"));
      bpmnMap.getFlowNodes().put(new QName(emptyComponent.getId()), emptyComponent);

      // new sequence flow starts from empty component, so as to be sure the condition from the gateway is not lost
      bpmnMap.changeSequenceFlowTarget(outSequences.get(0), emptyComponent);
      bpmnMap.createSequenceFlow(emptyComponent, targetNode);


      Misc.fastDynamicLog(FlowLinkTransformer.LOG, Level.INFO,
              "Adding empty activity %s after merge gateway: %s (%s)",
              emptyComponent.getId(), gateway.getId(), gateway);
      return true;
    }

    // Add link
    // remove sequenceFlow
    bpmnMap.deleteSequenceFlow(sequenceToReplace);
    final LinkComponent linkComponent = new LinkComponent(linkSource, targetNode, null);
    bpmnMap.getLinks().add(linkComponent);
    JoinConditionComponent.addNewAndLink(bpmnMap, linkComponent);

    // if sourceNode has no more outgoing sequenceFlow, add a sequenceFlow link to end
    if (bpmnMap.getOutgoingSequenceFlows(sourceNode).isEmpty()) {
      if (bpmnMap.getEndEvents().size() != 1) {
        throw new TransformerException("Only process with one end event is supported.");
      }
      // create a parallel gateway before last node
      final TParallelGateway lastGateway = this.getOrAddEndParallelGateway(bpmnMap);

      // create sequenceFlow from the source of the link to the new gateway.
      bpmnMap.createSequenceFlow(sourceNode, lastGateway);
    }
    Misc.fastDynamicLog(FlowLinkTransformer.LOG, Level.INFO, "Creating link: %s and removing sequenceFlow %s: %s (%s) -> %s (%s)",
            linkComponent.getName(),
            sequenceToReplace.getId(),
            ((TFlowNode) sequenceToReplace.getSourceRef()).getId(),
            sequenceToReplace.getSourceRef().getClass().getSimpleName(),
            ((TFlowNode) sequenceToReplace.getTargetRef()).getId(),
            sequenceToReplace.getTargetRef().getClass().getSimpleName());
    return true;
  }
}
