/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.transformers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.TFlowNodeComparator;
import org.ow2.orchestra.bpmn2bpel.components.FlowComponent;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TGateway;
import org.ow2.orchestra.jaxb.bpmn.TParallelGateway;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.util.Misc;


/**
 * @author Guillaume Porcher
 */
public final class FlowTransformer {

  private static final Logger LOG = Logger.getLogger(FlowTransformer.class.getName());

  /**
   * Get and replace a flow
   *
   * @param bpmnMap
   * @return
   */
  public boolean replaceFlow(final BpmnMap bpmnMap) {
    Misc.fastDynamicLog(FlowTransformer.LOG, Level.INFO, "Trying to replace flow");
    final List<TFlowNode> flowNodes = this.getMaxFlowComponent(bpmnMap);
    if (flowNodes.isEmpty()) {
      return false;
    }
    final TFlowNode firstNode = flowNodes.get(0);

    // start gateway is the source of the first node incoming sequenceFlow
    final TParallelGateway startGateway =
            (TParallelGateway) bpmnMap.getIncomingSequenceFlows(firstNode).iterator().next().getSourceRef();
    // endGateway id the target of the first node outgoing sequenceFlow
    final TParallelGateway endGateway =
            (TParallelGateway) bpmnMap.getOutgoingSequenceFlows(firstNode).iterator().next().getTargetRef();

    if (flowNodes.size() == 1) {
      // only one activity in the flow, don't replace it with a component.
      // check if there are sequenceFlows directly from startGateway to endGateway
      // and remove them
      for (final TFlowNode startNode : bpmnMap.getFlowNodes().values()) {
        if (startNode instanceof TParallelGateway) {
          final Set<TGateway> nodes = new HashSet<TGateway>();
          final Collection<TSequenceFlow> outgoings = bpmnMap.getOutgoingSequenceFlows(startNode);
          for (final TSequenceFlow seq : outgoings) {
            final TFlowNode target = (TFlowNode) seq.getTargetRef();
            if (!(target instanceof TParallelGateway)) {
              final List<TSequenceFlow> targetOut = bpmnMap.getOutgoingSequenceFlows(target);
              if (targetOut.size() == 1) {
                final TSequenceFlow targetSeq = targetOut.get(0);
                final TFlowNode nextNext = (TFlowNode) targetSeq.getTargetRef();
                if (nextNext instanceof TParallelGateway) {
                  nodes.add((TGateway) nextNext);
                }
              }
            }
          }
          // remove direct links from gateway to other gateway in a flow
          for (final TSequenceFlow seq : outgoings) {
            final TFlowNode target = (TFlowNode) seq.getTargetRef();
            if (nodes.contains(target)) {
              bpmnMap.deleteSequenceFlow(seq);
              Misc.fastDynamicLog(FlowTransformer.LOG, Level.INFO, "Removing unneeded sequenceFlow %s: %s (%s) -> %s (%s)",
                      seq.getId(),
                      ((TFlowNode) seq.getSourceRef()).getId(),
                      seq.getSourceRef().getClass().getSimpleName(),
                      ((TFlowNode) seq.getTargetRef()).getId(),
                      seq.getTargetRef().getClass().getSimpleName());
              return true;
            }
          }
        }
      }
      // if no direct link has been removed, nothing has been changed by this method.
      return false;
    } else {
      // flowNodes is a valid flow, replace it with a component
      final FlowComponent flowComponent = new FlowComponent(flowNodes, bpmnMap);
      flowComponent.setId(Misc.getHumanReadableId("component"));
      bpmnMap.getFlowNodes().put(new QName(flowComponent.getId()), flowComponent);

      // remove nodes
      // Replace first sequenceFlow
      bpmnMap.createSequenceFlow(startGateway, flowComponent);
      // replace last sequence flow
      bpmnMap.createSequenceFlow(flowComponent, endGateway);
      // remove flow nodes from list
      for (final TFlowNode node : flowNodes) {
        bpmnMap.removeNode(node);
      }

      // check if gateways can be removed is done in a separate transformer.
      Misc.fastDynamicLog(FlowTransformer.LOG, Level.INFO, "Creating flow component: %s", flowComponent);
      return true;
    }
  }

  private List<TFlowNode> getMaxFlowComponent(final BpmnMap bpmnMap) {
    List<TFlowNode> flowNodes = Collections.emptyList();
    for (final TFlowNode flowNode : bpmnMap.getFlowNodes().values()) {
      if (flowNode instanceof TParallelGateway) {
        final List<TFlowNode> r = this.getFlow(bpmnMap, flowNode);
        if (r.size() > flowNodes.size()) {
          flowNodes = r;
        }
      }
    }
    return flowNodes;
  }

  private List<TFlowNode> getFlow(final BpmnMap bpmnMap, final TFlowNode startNode) {
    if (!(startNode instanceof TParallelGateway)) {
      return Collections.emptyList();
    }
    final Map<TGateway, List<TFlowNode>> nodes = new TreeMap<TGateway, List<TFlowNode>>(TFlowNodeComparator.INSTANCE);
    final Collection<TSequenceFlow> outgoings = bpmnMap.getOutgoingSequenceFlows(startNode);
    for (final TSequenceFlow seq : outgoings) {
      final TFlowNode target = (TFlowNode) seq.getTargetRef();
      if (target != null && !(target instanceof TParallelGateway)) {
        final Collection<TSequenceFlow> targetOut = bpmnMap.getOutgoingSequenceFlows(target);
        if (targetOut.size() == 1) {
          final TSequenceFlow targetSeq = targetOut.iterator().next();
          final TFlowNode nextNext = (TFlowNode) targetSeq.getTargetRef();
          if (nextNext instanceof TParallelGateway) {
            if (!nodes.containsKey(nextNext)) {
              nodes.put((TParallelGateway) nextNext, new ArrayList<TFlowNode>());
            }
            nodes.get(nextNext).add(target);
          }
        }
      }
    }
    List<TFlowNode> flowNodes = Collections.emptyList();
    for (final List<TFlowNode> l : nodes.values()) {
      if (l.size() > flowNodes.size()) {
        flowNodes = l;
      }
    }
    return flowNodes;
  }

}
