/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.transformers;

import java.util.List;
import java.util.ListIterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpmn.TExclusiveGateway;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TGateway;
import org.ow2.orchestra.jaxb.bpmn.TParallelGateway;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.util.Misc;

/**
 * @author Guillaume Porcher
 */
public final class GatewayTransformer {

  private static final Logger LOG = Logger.getLogger(GatewayTransformer.class.getName());

  /**
   * Replace gateways so that: - gateways can only be in gateways or out
   * gateways, not bidirectionnal - gateways with one input and one output are
   * removed - gateways are merged when possible
   *
   * @param bpmnMap
   * @return true if something was changed in the process structure.
   */
  public boolean replaceGateways(final BpmnMap bpmnMap) {
    Misc.fastDynamicLog(GatewayTransformer.LOG, Level.INFO, "Trying to replace gateway");
    // get gateways
    for (final TFlowNode node : bpmnMap.getFlowNodes().values()) {
      if (node instanceof TGateway) {
        final TGateway oldGateway = (TGateway) node;
        final List<TSequenceFlow> incomingSequenceFlows = bpmnMap.getIncomingSequenceFlows(node);
        final List<TSequenceFlow> outgoingSequenceFlows = bpmnMap.getOutgoingSequenceFlows(node);
        // if gateway is bidirectional, split it in two gateways
        if (this.splitBidirectionalGateway(bpmnMap, oldGateway, incomingSequenceFlows, outgoingSequenceFlows)) {
          return true;
        }
        // if gateway performs no routing (one input and one output), remove it
        if (this.removeNoRoutingGateway(bpmnMap, oldGateway, incomingSequenceFlows, outgoingSequenceFlows)) {
          return true;
        }
        // if gateway is converging, try to merge it with the next gateway
        if (this.mergeWithNextGateway(bpmnMap, oldGateway, outgoingSequenceFlows)) {
          return true;
        }
        // if gateway is a diverging parallel gateway, try to merge it with the previous parallel diverging gateway.
        if (this.mergeWithPreviousGateway(bpmnMap, oldGateway, incomingSequenceFlows)) {
          return true;
        }
      }
    }
    return false;
  }

  private boolean mergeWithNextGateway(
          final BpmnMap bpmnMap,
          final TGateway oldGateway,
          final List<TSequenceFlow> outgoingSequenceFlows) {
    // if oldGateway is a converging gateway (one outgoing sequenceFlow)
    if (outgoingSequenceFlows.size() == 1) {
      // if next activity is a gateway converging of the same type, merge
      final TFlowNode nextNode = (TFlowNode) outgoingSequenceFlows.get(0).getTargetRef();
      if (oldGateway.getClass().isInstance(nextNode) && bpmnMap.getOutgoingSequenceFlows(nextNode).size() == 1) {
        // merge the two nodes
        for (final TSequenceFlow inSequenceFlow : bpmnMap.getIncomingSequenceFlows(oldGateway)) {
          bpmnMap.changeSequenceFlowTarget(inSequenceFlow, nextNode);
        }
        // remove old gateway
        bpmnMap.removeNode(oldGateway);

        Misc.fastDynamicLog(GatewayTransformer.LOG, Level.INFO, "Merging join gateway: %s (%s) and %s (%s)",
                oldGateway.getId(),
                oldGateway.getClass().getSimpleName(),
                nextNode.getId(),
                nextNode.getClass().getSimpleName());
        return true;
      }
    }
    return false;
  }

  private boolean mergeWithPreviousGateway(
          final BpmnMap bpmnMap,
          final TGateway oldGateway,
          final List<TSequenceFlow> incomingSequenceFlows) {
    // if oldGateway is a diverging parallel gateway
    if ((oldGateway instanceof TParallelGateway) && incomingSequenceFlows.size() == 1) {
      // if previous activity is a gateway of the same type, merge
      final TFlowNode previousNode = (TFlowNode) incomingSequenceFlows.get(0).getSourceRef();
      if ((previousNode instanceof TParallelGateway) && bpmnMap.getIncomingSequenceFlows(previousNode).size() == 1) {
        // merge the two nodes
        for (final TSequenceFlow outSequenceFlow : bpmnMap.getOutgoingSequenceFlows(oldGateway)) {
          bpmnMap.changeSequenceFlowSource(outSequenceFlow, previousNode);
        }
        // remove old gateway
        bpmnMap.removeNode(oldGateway);
        Misc.fastDynamicLog(GatewayTransformer.LOG, Level.INFO, "Merging split gateway: %s (%s) and %s (%s)",
                oldGateway.getId(),
                oldGateway.getClass().getSimpleName(),
                previousNode.getId(),
                previousNode.getClass().getSimpleName());

        return true;
      }
    }
    return false;
  }

  private boolean removeNoRoutingGateway(
          final BpmnMap bpmnMap,
          final TGateway oldGateway,
          final List<TSequenceFlow> incomingSequenceFlows,
          final List<TSequenceFlow> outgoingSequenceFlows) {
    if (incomingSequenceFlows.size() == 1 && outgoingSequenceFlows.size() == 1) {
      // remove gateway with only one incoming and one outgoing transition
      final TSequenceFlow incomingSequenceFlow = incomingSequenceFlows.get(0);
      final TSequenceFlow outgoingSequenceFlow = outgoingSequenceFlows.get(0);

      // replace target of sequenceFlow
      final TFlowNode targetNode = (TFlowNode) outgoingSequenceFlow.getTargetRef();
      bpmnMap.changeSequenceFlowTarget(incomingSequenceFlow, targetNode);

      // replace condition
      if (outgoingSequenceFlow.getConditionExpression() != null) {
        if (incomingSequenceFlow.getConditionExpression() == null) {
          incomingSequenceFlow.setConditionExpression(outgoingSequenceFlow.getConditionExpression());
        } else {
          String condition = "(";
          ListIterator<Object> iterator = incomingSequenceFlow.getConditionExpression().getContent().listIterator();
          while (iterator.hasNext()) {
            final Object o = iterator.next();
            if (o instanceof String) {
              condition += o;
            }
          }
          condition += ") and (";
          iterator = outgoingSequenceFlow.getConditionExpression().getContent().listIterator();
          while (iterator.hasNext()) {
            final Object o = iterator.next();
            if (o instanceof String) {
              condition += o;
            }
          }
          condition += ")";
          incomingSequenceFlow.getConditionExpression().getContent().clear();
          incomingSequenceFlow.getConditionExpression().getContent().add(condition);
        }
      }

      // remove outgoingSequenceFlow
      bpmnMap.deleteSequenceFlow(outgoingSequenceFlow);
      // remove gateway
      bpmnMap.removeNode(oldGateway);
      Misc.fastDynamicLog(GatewayTransformer.LOG, Level.INFO, "Removing gateway with one input and one output: %s (%s)",
              oldGateway.getId(),
              oldGateway.getClass().getSimpleName());

      return true;
    }
    return false;
  }

  private boolean splitBidirectionalGateway(
          final BpmnMap bpmnMap,
          final TGateway oldGateway,
          final List<TSequenceFlow> incomingSequenceFlows,
          final List<TSequenceFlow> outgoingSequenceFlows) {
    if (incomingSequenceFlows.size() > 1 && outgoingSequenceFlows.size() > 1) {
      // split gateway in fork only and join only gateways
      final TGateway newGateway;
      if (oldGateway instanceof TParallelGateway) {
        newGateway = new TParallelGateway();
      } else if (oldGateway instanceof TExclusiveGateway) {
        newGateway = new TExclusiveGateway();
      } else {
        throw new TransformerException("Bidirectionnal gateway " + oldGateway.getClass().getSimpleName() + " not supported");
      }
      newGateway.setId(Misc.getHumanReadableId("component"));
      bpmnMap.getFlowNodes().put(new QName(newGateway.getId()), newGateway);

      for (final TSequenceFlow outgoingSequenceFlow : outgoingSequenceFlows) {
        bpmnMap.changeSequenceFlowSource(outgoingSequenceFlow, newGateway);
      }
      bpmnMap.createSequenceFlow(oldGateway, newGateway);


      Misc.fastDynamicLog(GatewayTransformer.LOG, Level.INFO,
              "Spliting bidirectionnal gateway: %s (%s) to ",
              oldGateway.getId(), oldGateway, newGateway.getId(), newGateway);
      return true;
    }
    return false;
  }
}
