/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.transformers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.components.SequenceComponent;
import org.ow2.orchestra.jaxb.bpmn.TEndEvent;
import org.ow2.orchestra.jaxb.bpmn.TEventBasedGateway;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TGateway;
import org.ow2.orchestra.jaxb.bpmn.TIntermediateCatchEvent;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.jaxb.bpmn.TStartEvent;
import org.ow2.orchestra.util.Misc;


/**
 * @author Guillaume Porcher
 */
public final class SequenceTransformer {

  private static final Logger LOG = Logger.getLogger(SequenceTransformer.class.getName());

  /**
   * Get and replace max sequence.
   * Does not replace the start and end event activities
   *
   * @param bpmnMap
   * @return true if something was replaced
   */
  public boolean replaceMaxSequence(final BpmnMap bpmnMap) {
    Misc.fastDynamicLog(SequenceTransformer.LOG, Level.INFO, "Trying to replace sequence");
    List<TFlowNode> maxSequenceNodes = Collections.emptyList();
    for (final TFlowNode flowNode : bpmnMap.getFlowNodes().values()) {
      final List<TFlowNode> sequenceNodes = this.getSequence(bpmnMap, flowNode);
      if (sequenceNodes.size() > maxSequenceNodes.size()) {
        maxSequenceNodes = sequenceNodes;
      }
    }
    // startEvent and endEvent can only be included in the same sequence.
    // start and endEvents are used for links generation and should not be removed from the process before the last step.
    if (maxSequenceNodes.size() > 1) {
      final TFlowNode firstNode = maxSequenceNodes.get(0);
      final TFlowNode lastNode = maxSequenceNodes.get(maxSequenceNodes.size() - 1);
      if (firstNode instanceof TStartEvent) {
        if (!(lastNode instanceof TEndEvent)) {
          maxSequenceNodes.remove(0);
        }
      } else if (lastNode instanceof TEndEvent) {
        maxSequenceNodes.remove(maxSequenceNodes.size() - 1);
      }
    }
    // if sequence contains only one element, it is not a sequence
    if (maxSequenceNodes.size() < 2) {
      return false;
    }
    // maxSequenceNodes is a valid sequence, replace it with a component
    final SequenceComponent sequenceComponent = new SequenceComponent(maxSequenceNodes, bpmnMap);
    sequenceComponent.setId(Misc.getHumanReadableId("component"));
    bpmnMap.getFlowNodes().put(new QName(sequenceComponent.getId()), sequenceComponent);

    // Replace first sequenceFlow
    final Collection<TSequenceFlow> inSeqFlows = bpmnMap.getIncomingSequenceFlows(maxSequenceNodes.get(0));
    for (final TSequenceFlow inSeqFlow : inSeqFlows) {
      bpmnMap.createSequenceFlow((TFlowNode) inSeqFlow.getSourceRef(), sequenceComponent);
      bpmnMap.deleteSequenceFlow(inSeqFlow);
    }
    // replace last sequence flow
    final Collection<TSequenceFlow> outSeqFlows =
            bpmnMap.getOutgoingSequenceFlows(maxSequenceNodes.get(maxSequenceNodes.size() - 1));
    for (final TSequenceFlow outSeqFlow : outSeqFlows) {
      bpmnMap.createSequenceFlow(sequenceComponent, (TFlowNode) outSeqFlow.getTargetRef());
      bpmnMap.deleteSequenceFlow(outSeqFlow);
    }
    // remove sequence nodes from list
    for (final TFlowNode node : maxSequenceNodes) {
      bpmnMap.removeNode(node);
    }

    Misc.fastDynamicLog(SequenceTransformer.LOG, Level.INFO, "Creating sequence component: %s", sequenceComponent);

    return true;
  }

  /**
   * Find the longest sequence of activities after the startNode activity.
   *
   * @param bpmnMap
   * @param startNode
   * @return
   */
  protected List<TFlowNode> getSequence(final BpmnMap bpmnMap, final TFlowNode startNode) {
    TFlowNode nextNode = startNode;
    final List<TFlowNode> sequenceNodes = new ArrayList<TFlowNode>();

    // For Pick structures, don't add a messageEvent following a event based gateway to a sequence.
    if (startNode instanceof TIntermediateCatchEvent) {
      final Object previousNode = bpmnMap.getIncomingSequenceFlows(startNode).get(0).getSourceRef();
      if (previousNode instanceof TEventBasedGateway) {
        return sequenceNodes;
      }
    }
    while (nextNode != null && !(nextNode instanceof TGateway)) {
      sequenceNodes.add(nextNode);
      final Collection<TSequenceFlow> outgoings = bpmnMap.getOutgoingSequenceFlows(nextNode);
      if (outgoings.size() != 1) {
        nextNode = null;
      } else {
        final TSequenceFlow firstSeq = outgoings.iterator().next();
        nextNode = (TFlowNode) firstSeq.getTargetRef();
      }
    }
    return sequenceNodes;
  }

}
