/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.transformers;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.ListIterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.components.WhileComponent;
import org.ow2.orchestra.jaxb.bpmn.TExclusiveGateway;
import org.ow2.orchestra.jaxb.bpmn.TExpression;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.util.Misc;


/**
 * @author Guillaume Porcher
 */
public final class WhileTransformer {

  private static final Logger LOG = Logger.getLogger(WhileTransformer.class.getName());

  /**
   * Get and replace a while
   *
   * @param bpmnMap
   * @return
   */
  public boolean replaceWhile(final BpmnMap bpmnMap) {
    Misc.fastDynamicLog(WhileTransformer.LOG, Level.INFO, "Trying to replace while");
    // find first while component
    List<TFlowNode> flowNodes = Collections.emptyList();
    for (final TFlowNode flowNode : bpmnMap.getFlowNodes().values()) {
      if (flowNode instanceof TExclusiveGateway) {
        flowNodes = this.getWhile(bpmnMap, flowNode);
        if (!flowNodes.isEmpty()) {
          break;
        }
      }
    }
    if (flowNodes.isEmpty()) {
      return false;
    }

    // get activities
    final TExclusiveGateway startGateway = (TExclusiveGateway) flowNodes.get(0);
    final TExclusiveGateway endGateway = (TExclusiveGateway) flowNodes.get(1);
    final TFlowNode componentNode = flowNodes.get(2);

    // get condition
    TExpression condition = null;
    for (final TSequenceFlow endGatewaySeq : bpmnMap.getOutgoingSequenceFlows(endGateway)) {
      if (!endGatewaySeq.equals(endGateway.getDefault())) {
        condition = endGatewaySeq.getConditionExpression();
        if (!endGatewaySeq.getTargetRef().equals(componentNode)) {
          final ListIterator<Object> iterator = condition.getContent().listIterator();
          while (iterator.hasNext()) {
            final Object o = iterator.next();
            if (o instanceof String) {
              final String expr = (String) o;
              iterator.remove();
              iterator.add("not(" + expr + ")");
            }
          }
        }
      }
    }

    // flowNodes is a valid while, replace it with a component
    final WhileComponent whileComponent = new WhileComponent(componentNode, condition);
    whileComponent.setId(Misc.getHumanReadableId("component"));
    bpmnMap.getFlowNodes().put(new QName(whileComponent.getId()), whileComponent);

    // remove nodes
    // Replace first sequenceFlow
    for (final TSequenceFlow startGatewayIn : bpmnMap.getIncomingSequenceFlows(startGateway)) {
      if (!startGatewayIn.getSourceRef().equals(componentNode)) {
        // startGatewayIn is first sequenceFlow
        bpmnMap.createSequenceFlow((TFlowNode) startGatewayIn.getSourceRef(), whileComponent);
      }
    }
    // replace last sequence flow
    for (final TSequenceFlow endGatewayOut : bpmnMap.getOutgoingSequenceFlows(endGateway)) {
      if (!endGatewayOut.getTargetRef().equals(componentNode)) {
        // endGatewayOut is last sequenceFlow
        bpmnMap.createSequenceFlow(whileComponent, (TFlowNode) endGatewayOut.getTargetRef());
      }
    }

    // remove sequence nodes from list
    for (final TFlowNode node : flowNodes) {
      bpmnMap.removeNode(node);
    }

    Misc.fastDynamicLog(WhileTransformer.LOG, Level.INFO, "Creating while component: %s (remove gateways: %s and %s)",
            whileComponent,
            startGateway.getId(),
            endGateway.getId());
    return true;
  }

  /**
   * Get while components:
   * While component is a XOR merge gateway with 1 outgoing transition and 2 incoming transitions
   * The outgoing transition goes to a XOR decision gateway with 2 outgoing transitions:
   * and one goes to a component, with one outgoing transition going to the first XOR merge gateway.
   *
   * @param bpmnMap
   * @param startNode
   * @return
   */
  private List<TFlowNode> getWhile(final BpmnMap bpmnMap, final TFlowNode startNode) {
    // is startNode the XOR merge gateway ?
    if (!(startNode instanceof TExclusiveGateway)) {
      return Collections.emptyList();
    }
    final List<TSequenceFlow> outgoings = bpmnMap.getOutgoingSequenceFlows(startNode);
    if (outgoings.size() != 1) {
      return Collections.emptyList();
    }
    // get unique outgoing transition
    final TSequenceFlow seq = outgoings.get(0);
    // get target node
    final TFlowNode target = (TFlowNode) seq.getTargetRef();
    if (target != null && (target instanceof TExclusiveGateway)) {
      // is target node a XOR decision gateway ?
      final Collection<TSequenceFlow> targetOuts = bpmnMap.getOutgoingSequenceFlows(target);
      if (targetOuts.size() == 2) {
        TFlowNode whileNode = null;
        // search for component going back to startNode
        for (final TSequenceFlow targetOut : targetOuts) {
          final TFlowNode nextNext = (TFlowNode) targetOut.getTargetRef();
          final Collection<TSequenceFlow> nextTargetOuts = bpmnMap.getOutgoingSequenceFlows(nextNext);
          if (nextTargetOuts.size() == 1
                  && nextTargetOuts.iterator().next().getTargetRef().equals(startNode)) {
            // nextNext is the component and we have found a WHILE structure.
            whileNode = nextNext;
          }
        }
        final TExclusiveGateway endGateway = (TExclusiveGateway) target;
        if (whileNode != null && endGateway.getDefault() != null) {
          return Arrays.asList(startNode, endGateway, whileNode);
        }
      }
    }
    return Collections.emptyList();
  }

}
